@@ -1,5 +1,5 @@
 /*
- * $Id: ODBC.h,v 1.4 1997/07/16 19:26:20 timbo Exp $
+ * $Id: ODBC.h,v 1.5 1997/07/25 11:50:07 timbo Exp $
  * Copyright (c) 1994,1995,1996,1997  Tim Bunce
  *
  * You may distribute under the terms of either the GNU General Public
@@ -14,6 +14,6 @@
 
 #include "dbdimp.h"
 
-#include <DBDXSI.h>	/* from DBI. Load this after mysql.h */
+#include <dbd_xsh.h>	/* from DBI. Load this after mysql.h */
 
 /* end of ODBC.h */
